/* Modifications to upstart to allow passing of the state of currently
 * running jobs when upstart re-execs itself.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */


#include <sys/types.h>
#include <string.h>
#include <nih/string.h>
#include <nih/logging.h>
#include <nih/alloc.h>
#include "job.h"
#include "restore.h"


/**
 * store_pid:
 * @proc:    pointer to a JobProcess structure or NULL
 * @args:    array of arguments
 * @parent:  parent of @args
 * @len:     current length of @args
 * 
 * Does nothing if @args is null.
 * Pushes "X" onto @args if @args is non-null but @process is null.
 * Otherwise: extracts the pid from @process, creates a hexadecimal string
 * form of it, pushes a copy onto the array @args, then frees the original.
 */

void store_pid(JobProcess *proc, char ***args, const void *parent, size_t *len)
{
	char *pidstr;
	if (! *args) return;
	pidstr = (proc ? nih_sprintf(NULL, "%x", proc->pid) : "X");

	if (! pidstr) {
		nih_error("Unable to allocate storage to save process id");
		*args = NULL;
		return;
	}
	if (! nih_str_array_add(args, parent, len, pidstr)) {
		nih_error("Unable to add process id to argument list");
		*args = NULL;
	}

	if (proc) nih_free(pidstr);
}
 
/**
 * unpack_pid:
 * @pidstr: hexadecimal string created by pack_pid
 * @err: pointer to flag to set if error occurs
 *
 * Returns the process id corresponding to the string passed in pidstr.
 * If an error occurs, returns 0 and sets the flag pointed to by @err.
 **/

pid_t unpack_pid(char *pidstr, int *err) {
	pid_t result = 0;
	while (*pidstr) {
		unsigned char c = *pidstr++;
		result <<= 4;
		if (c >= '0' && c <= '9') result += (c-'0');
		else if (c >= 'a' && c <= 'f') result += 10 + (c-'a');
		else {
			*err = 1;
			return 0;
		}
	}
	return result;
}

/**
 * retrieve_pid:
 * @arg: 	pointer to array of arguments from which to read pid string
 * @jobname: 	name of job whose pids are being retrieved
 * @err: 	pointer to flag to set if error occurs
 * @pid: 	pointer to location to hold the retrieved process id
 * @present:	pointer to a location to hold 1 if the process id was present
 *
 * Reads a process id or no-process-present indicator from the hexadecimal
 * string pointed to by @arg, increments @arg to point to the next string in
 * the array unless it already points to a null string, and stores into
 * @present a value of 0 or 1 indicating whether or not a process id is
 * present, and stores into @pid the unpacked process id (or zero if there
 * is no process present or an error occurred). If an error occured, sets
 * the flag pointed to by @err and printes an error message.
 **/

void retrieve_pid(char ***arg, char *jobname, int *err,
						pid_t *pid, int *present)
{
	*pid = 0;
	*present = 0;
	char *str;
	if (! arg || ! *arg || ! **arg) {
		nih_error("Ran out of arguments when restoring job %s",
			  jobname);
		*err = 1;
		return;
	}
	str = **arg;
	if (str) ++(*arg);
	if (strcmp(str, "X")==0) return;
	*present = 1;
	*pid = unpack_pid(str, err);
	if (*err) nih_error("Cannot parse pid %s for job %s", str, jobname);
}

/**
 * restore_pid:
 * @proc: 	pointer to a JobProcess structure, or NULL
 * @pid: 	process id
 * @present: 	integer which is 1 if process id is applicable
 * @jobname:	name of job whose process ids are being restored
 *
 * Checks that the non-null status of @proc matches the true/false
 * status of @present. If there is a mismatch, prints an error message.
 * If @present is true and @proc is non-null, stores the process id from
 * @pid into the JobProcess structure pointed to by @proc.
 */

void restore_pid(JobProcess *proc, pid_t pid, int present, char *jobname)
{
	if (proc == NULL && ! present) return;
	if (proc != NULL && present) {
		proc->pid = pid;
		return;
	}
	nih_error("Unable to restore process id for job %s", jobname);
}

/**
 * save_current_configuration:
 * @args:   array of arguments to pass to new init
 * @parent: parent of @args
 * @len:    current length of @args
 *
 * Pushes, onto the array @args, information about currently running
 * jobs. Each job consists of seven arguments:
 *   (1) The job name
 *   (2) A 3-character state/goal string:
 *         First character for the goal: 0=goal JOB_STOP, 1=goal JOB_START
 *         Second character for the current state:
 *             W=JOB_WAITING,
 *             S=JOB_STARTING, A=JOB_PRE_START, B=JOB_SPAWNED, C=JOB_POST_START,
 *             R=JOB_RUNNING,
 *             T=JOB_STOPPING, X=JOB_PRE_STOP, Y=JOB_KILLED, Z=JOB_POST_STOP
 *         Third character: N=normal job, I=instance job
 *   (3) The associated PROCESS_MAIN process id
 *   (4) The associated PROCESS_PRE_START process id
 *   (5) The associated PROCESS_POST_START process id
 *   (6) The associated PROCESS_PRE_STOP process id
 *   (7) The associated PROCESS_POST_STOP process id
 * Process ids are in hexadecimal using lower-case letters (0-9,a-f),
 * or 0 if there is a JobProcess structure but no running process, or
 * "X" if there is no JobProcess structure.
 *
 */
 
void save_current_configuration(char ***args, const void *parent, size_t *len)
{
	NIH_HASH_FOREACH(jobs, iter) {
		Job *job = (Job *) iter;
		char goal_state[4];
		JobProcess **pids;

		
		if (job->replacement_for != NULL) continue;
		if (job->state == JOB_DELETED) continue;
		if (job->instance && ! job->instance_of) continue;
		
		if (! nih_str_array_add(args, parent, len, job->name))
			return;
		
		if (job->goal == JOB_START) goal_state[0] = '1';
		else goal_state[0] = '0';
		
		switch (job->state)
			{
			case JOB_WAITING:    goal_state[1] = 'W'; break;
			case JOB_STARTING:   goal_state[1] = 'S'; break;
			case JOB_PRE_START:  goal_state[1] = 'A'; break;
			case JOB_SPAWNED:    goal_state[1] = 'B'; break;
			case JOB_POST_START: goal_state[1] = 'C'; break;
			case JOB_RUNNING:    goal_state[1] = 'R'; break;
			case JOB_STOPPING:   goal_state[1] = 'T'; break;
			case JOB_PRE_STOP:   goal_state[1] = 'X'; break;
			case JOB_KILLED:     goal_state[1] = 'Y'; break;
			case JOB_POST_STOP:  goal_state[1] = 'Z'; break;
			default:             goal_state[1] = '?';      
			}
		if (job->instance_of) goal_state[2] = 'I';
		else goal_state[2] = 'N';
		goal_state[3] = 0;
		
		if (! nih_str_array_add(args, parent, len, goal_state))
			return;
		
		pids = job->process;
		store_pid(pids[PROCESS_MAIN],       args, parent, len);
		store_pid(pids[PROCESS_PRE_START],  args, parent, len);
		store_pid(pids[PROCESS_POST_START], args, parent, len);
		store_pid(pids[PROCESS_PRE_STOP],   args, parent, len);
		store_pid(pids[PROCESS_POST_STOP],  args, parent, len);
	}
}

/**
 * restore_job_state:
 * @jobname: name of job whose state we are to restore
 * @goal_state: the goal/state string generated by save_current_configuration
 * @pids: array of five process ids 
 * @presents: array of five integers indicating which process ids are applicable
 */

void restore_job_state (char *jobname, char *goal_state,
				pid_t *pids, int *presents)
{
	Job *job;
	JobGoal goal = JOB_STOP;
	JobState state = JOB_WAITING;
	int is_instance = 0;
	int is_invalid = 0;
	JobProcess **procs;

	if (jobname == NULL) {
		nih_error("Missing name for restored job");
		return;
	}

	if (goal_state == NULL) {
		nih_error("Missing goal/state for restored job %s", jobname);
		return;
	}
	
	if (strlen(goal_state) != 3) {
		nih_error("Invalid length goal/state %s for restored job %s",
			  goal_state, jobname);
		return;
	}

	switch (goal_state[0])
		{
		case '0': goal = JOB_STOP; break;
		case '1': goal = JOB_START; break;
		default: is_invalid = 1; break;
		}
	
	switch (goal_state[1])
		{
		case 'W': state = JOB_WAITING; break;
		case 'S': state = JOB_STARTING; break;
		case 'A': state = JOB_PRE_START; break;
		case 'B': state = JOB_SPAWNED; break;
		case 'C': state = JOB_POST_START; break;
		case 'R': state = JOB_RUNNING; break;
		case 'T': state = JOB_STOPPING; break;
		case 'X': state = JOB_PRE_STOP; break;
		case 'Y': state = JOB_KILLED; break;
		case 'Z': state = JOB_POST_STOP; break;
		default: is_invalid = 1;
		}

	switch (goal_state[2])
		{
		case 'N': break;
		case 'I': is_instance = 1; break;
		default: is_invalid = 1;
		}
	
	if (is_invalid) {
		nih_error("Invalid goal/state %s for restored job %s",
			  goal_state, jobname);
		return;
	}
	job = job_find_by_name(jobname);
	if (! job) {
		nih_error("Unknown job %s passed to re-exec", jobname);
		return;
	}				
	if (is_instance) {
		if (! job->instance) {
			nih_error("Can not restore job %s as instance job",
				  jobname);
			return;
		}
		job = job_instance(job);
	} else {
		if (job->instance) {
			nih_error("Can not restore job %s as non-instance job",
				  jobname);
			return;
		}
	}
	job->goal = goal;
	job->state = state;
	procs=job->process;
	restore_pid(procs[PROCESS_MAIN],       pids[0], presents[0], jobname);
	restore_pid(procs[PROCESS_PRE_START],  pids[1], presents[1], jobname);
	restore_pid(procs[PROCESS_POST_START], pids[2], presents[2], jobname);
	restore_pid(procs[PROCESS_PRE_STOP],   pids[3], presents[3], jobname);
	restore_pid(procs[PROCESS_POST_STOP],  pids[4], presents[4], jobname);
}

/**
 * restore_current_configuration:
 * @args:   non-option arguments that were passed to the new init
 */

void restore_configuration(char **args)
{
	if (! args) return;
	
	char **arg = args;
	while (arg && *arg) {
		char *jobname = *(arg++);
		char *goal_state = *arg;
		pid_t pids[5];
		int presents[5];
		int is_invalid = 0;

		if (*arg) ++arg;
		for (int i = 0; i < 5; i++) 
			retrieve_pid(&arg, jobname, &is_invalid,
					&pids[i], &presents[i]);

		if (! is_invalid)
			restore_job_state (jobname, goal_state, pids, presents);
	}
}
